# Standard library imports
import logging
import os
import sys
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import cast

# Third party imports
from rich.console import Console
from rich.logging import RichHandler
from rich.progress import Progress, SpinnerColumn, TextColumn

# Project imports
from mindgard.auth import load_access_token, login, logout
from mindgard.constants import API_BASE, DASHBOARD_URL, VERSION
from mindgard.dataset_generation import create_custom_dataset
from mindgard.exceptions import MGException
from mindgard.external_model_handlers.llm_model import llm_message_handler
from mindgard.orchestrator import OrchestratorSetupRequest
from mindgard.preflight import preflight_llm
from mindgard.run_functions.external_models import (
    model_test_output_factory,
    model_test_polling,
    model_test_submit_factory,
)
from mindgard.run_functions.list_tests import list_test_output, list_test_polling, list_test_submit
from mindgard.run_functions.sandbox_test import submit_sandbox_polling, submit_sandbox_submit_factory
from mindgard.run_poll_display import cli_run
from mindgard.utils import (
    CliResponse,
    ConflictError,
    conflicting_args_error,
    convert_test_to_cli_response,
    invalid_project_id_error,
    is_version_outdated,
    parse_toml_and_args_into_final_args,
    print_to_stderr,
    unspecified_project_id_error,
)
from mindgard.wrappers.llm import LLMModelWrapper
from mindgard.wrappers.utils import parse_args_into_model

from .api.attack_multiturn import AttackMultiturnClient
from .api.exceptions import ClientException
from .cli_args import parse_args
from .multiturn.attack import MultiturnAttack
from .project.project import create_project, list_projects
from .recon.constants import recon_constants
from .recon.run_guardrail import run_guardrail_recon
from .recon.run_recon import run_recon

debug_help = lambda: print(
    "\033[93mTry running with `mindgard --log-level=debug ...` for more information, or ` --log-file /path/to/file.log` after your command to save output to disk.\033[0m"
)


def run_cli() -> None:

    args = parse_args(sys.argv[1:])
    log_handlers = [
        RichHandler(
            console=Console(stderr=True), locals_max_string=None, locals_max_length=None, level=args.log_level.upper()
        )
    ]
    if args.log_file:
        log_path = args.log_file.expanduser().resolve()
        file_formatter = logging.Formatter(
            fmt="%(asctime)s.%(msecs)03d %(levelname)-8s %(name)-12s[%(process)d] %(threadName)-12s %(pathname)s:%(lineno)d %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler = logging.FileHandler(log_path, mode="w")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(file_formatter)
        log_handlers.append(file_handler)

    FORMAT = "%(asctime)s.%(msecs)02d; %(message)s"
    logging.basicConfig(
        level=logging.DEBUG,  # always do debug and then filter at the handlers level (file log is always debug)
        format=FORMAT,
        datefmt="[%X]",
        handlers=log_handlers,
    )

    if not (sys.version_info.major == 3 and sys.version_info.minor >= 10):
        print_to_stderr("Python 3.10 or later is required to run the Mindgard CLI.")
        sys.exit(2)

    if new_version := is_version_outdated():
        print_to_stderr(
            f"New version available: {new_version}. Please upgrade as older versions of the CLI may not be actively maintained."
        )

    logging.debug(f"Mindgard version: {VERSION}")

    if args.command == "login":
        login(instance=args.instance)
    elif args.command == "logout":
        logout()
    elif args.command == "list":
        if args.list_command == "projects":
            list_projects()
            exit(CliResponse(0).code())
        else:
            print_to_stderr("Provide a resource to list. Eg `list projects`.")
    elif args.command == "sandbox":
        submit_sandbox_submit = submit_sandbox_submit_factory()
        submit_sandbox_output = model_test_output_factory(risk_threshold=100)

        cli_response = cli_run(
            submit_func=submit_sandbox_submit,
            polling_func=submit_sandbox_polling,
            output_func=submit_sandbox_output,
            json_out=args.json,
        )
        exit(convert_test_to_cli_response(test=cli_response, risk_threshold=100).code())
    elif args.command == "create":
        if args.create_command == "dataset":
            create_custom_dataset(args)
        elif args.create_command == "project":
            create_project(args)

        else:
            print("Unknown create command. Please see `mindgard create --help` for more information.")

    elif args.command == "validate":
        console = Console(highlight=False)  # Disabled highlighting to avoid styling issues with certain terminals
        final_args = {}
        try:
            final_args = parse_toml_and_args_into_final_args(args.config_file, args)
        except ConflictError as e:
            conflicting_args_error(console, e)
            exit(CliResponse(1).code())
        except ValueError as e:
            console.print(f"[red bold]{e}")
            exit(CliResponse(1).code())
        model_wrapper = parse_args_into_model(final_args)
        passed_preflight = preflight_llm(model_wrapper, console=console, json_out=final_args["json"])
        if not final_args["json"]:
            console.print(
                f"{'[green bold]Target contactable!' if passed_preflight else '[red bold]Target not contactable!'}"
            )

    elif args.command in ["test", "recon", "run"]:
        console = Console(highlight=False)  # Disabled highlighting to avoid styling issues with certain terminals
        final_args = {}
        try:
            final_args = parse_toml_and_args_into_final_args(args.config_file, args)
        except ConflictError as e:
            conflicting_args_error(console, e)
            exit(CliResponse(1).code())
        except ValueError as e:
            console.print(f"[red bold]{e}")
            exit(CliResponse(1).code())

        if not final_args["project_id"]:
            unspecified_project_id_error(console)
            exit(CliResponse(1).code())

        model_wrapper = parse_args_into_model(final_args)
        passed_preflight = preflight_llm(model_wrapper, console=console, json_out=final_args["json"])

        if not final_args["json"]:
            console.print(
                f"{'[green bold]Target contactable!' if passed_preflight else '[red bold]Target not contactable!'}"
            )

        if passed_preflight:
            match args.command:
                # Test command
                case "test":
                    if os.getenv("MINDGARD_TOGGLE_USE_LIB") == "true":
                        # Project imports
                        from mindgard.main_lib import run_test

                        run_test(final_args, model_wrapper)
                    else:
                        request = OrchestratorSetupRequest(
                            projectID=final_args["project_id"],
                            parallelism=final_args["parallelism"],
                            system_prompt=final_args["system_prompt"],
                            dataset=final_args.get("custom_dataset", final_args["dataset"]),
                            custom_dataset=final_args.get("custom_dataset", None),
                            modelType="llm",
                            attackSource="user",
                            attackPack=final_args["attack_pack"],
                            exclude=final_args["exclude"],
                            include=final_args["include"],
                            prompt_repeats=final_args["prompt_repeats"],
                        )
                        submit = model_test_submit_factory(
                            request=request,
                            model_wrapper=cast(LLMModelWrapper, model_wrapper),
                            message_handler=llm_message_handler,
                        )

                        output = model_test_output_factory(risk_threshold=int(final_args["risk_threshold"]))
                        cli_response = cli_run(
                            submit, model_test_polling, output_func=output, json_out=final_args["json"]
                        )
                        exit(
                            convert_test_to_cli_response(
                                test=cli_response, risk_threshold=int(final_args["risk_threshold"])
                            ).code()
                        )  # type: ignore

                # Reconnaissance commands
                case recon_constants.recon:
                    if args.recon_subcommand == recon_constants.guardrail:
                        run_guardrail_recon(console, model_wrapper, final_args)
                    elif args.recon_subcommand == recon_constants.input_encoding:
                        run_recon(
                            type="input-encoding", console=console, model_wrapper=model_wrapper, final_args=final_args
                        )
                    elif args.recon_subcommand == recon_constants.output_encoding:
                        run_recon(
                            type="output-encoding", console=console, model_wrapper=model_wrapper, final_args=final_args
                        )
                    elif args.recon_subcommand == recon_constants.non_contextual:
                        print_to_stderr("Non-contextual is not yet implemented.")
                    elif args.recon_subcommand == recon_constants.output_formatting:
                        run_recon(
                            type="output-formatting",
                            console=console,
                            model_wrapper=model_wrapper,
                            final_args=final_args,
                        )
                    elif args.recon_subcommand == recon_constants.output_rendering:
                        run_recon(
                            type="output-rendering",
                            console=console,
                            model_wrapper=model_wrapper,
                            final_args=final_args,
                        )
                    elif args.recon_subcommand == recon_constants.code_generation:
                        print_to_stderr("Code generation is not yet implemented.")
                    elif args.recon_subcommand == recon_constants.system_prompt_extraction:
                        run_recon(
                            type="system-prompt-extraction",
                            console=console,
                            model_wrapper=model_wrapper,
                            final_args=final_args,
                        )
                    elif args.recon_subcommand == recon_constants.tool_discovery:
                        run_recon(
                            type="tool-discovery", console=console, model_wrapper=model_wrapper, final_args=final_args
                        )
                    else:
                        print_to_stderr("Provide a recon sub-command. See: $ mindgard recon --help")

                # Run commands
                case "run":
                    if args.run_subcommands == "multiturn":
                        if len(final_args["include"]) != 1:
                            console.print("--include currently only supports one element")
                            exit(CliResponse(1).code())
                        access_token = load_access_token()
                        goal = final_args["goal"]
                        project_id = final_args["project_id"]
                        attack_name = final_args["include"][0]
                        mt_attack_client = AttackMultiturnClient(API_BASE, access_token, project_id)
                        multiturn_attack = MultiturnAttack(
                            call_system_under_test=model_wrapper,
                            ma_client=mt_attack_client,
                            project_id=project_id,
                            goal=goal,
                            attack_name=attack_name,
                        )

                        def run_attack():
                            start_response = multiturn_attack.start()
                            multiturn_attack.poll(start_response.run_id)
                            attack_results = multiturn_attack.get_results(start_response.run_id)
                            url = f"{DASHBOARD_URL}/results/multiturns/{attack_results.attack_session_ids[0]}?project_id={project_id}"
                            return url

                        url = None
                        with Progress(
                            SpinnerColumn(style="bold yellow"),
                            TextColumn(f"[bold yellow]Running Multiturn Attack: {attack_name}"),
                            transient=True,
                            console=console,
                        ) as progress:
                            task_id = progress.add_task("running", total=None)

                            with ThreadPoolExecutor(max_workers=1) as executor:
                                future = executor.submit(run_attack)

                                try:
                                    url = future.result()
                                except ClientException as e:
                                    raise e
                                except Exception as e:
                                    console.print(f"Error trying to run multiturn attack, {e}")
                                    exit(CliResponse(1).code())
                                progress.update(task_id, completed=100)
                        if not url:
                            console.print(f"Error did not get results url")
                            exit(CliResponse(1).code())
                        console.print(f"Multiturn Attack complete: {url}")
                        exit(CliResponse(0).code())
                case _:
                    print_to_stderr("Which command are you looking for? See: $ mindgard --help")

        else:
            exit(CliResponse(1).code())

    else:
        print_to_stderr("Which command are you looking for? See: $ mindgard --help")


def main() -> None:
    console = Console(stderr=True)
    try:
        run_cli()
    except ValueError as e:
        console.print(f"[red bold]{e}")
        invalid_project_id_error(console, str(e))
        exit(2)
    except MGException:
        debug_help()
        exit(2)
    except ClientException as e:
        console.print(f"[red bold]Error: {e.message}")
        invalid_project_id_error(console, str(e))
        exit(2)
    except Exception:
        traceback.print_exc()
        debug_help()
        exit(2)



# Standard library imports
import json
import os
from typing import Any, Dict, List, Literal, Optional, cast

# Third party imports
from pydantic import BaseModel, model_validator
from requests.exceptions import HTTPError

# Project imports
from mindgard.api_service import (
    MGHTTPError,
    api_get,
    api_post,
)
from mindgard.constants import API_BASE, DASHBOARD_URL
from mindgard.types import type_get_request_function, type_orchestrator_source, type_post_request_function

# Tells pydantic to stop worrying about model_type namespace collision
BaseModel.model_config["protected_namespaces"] = ()


class OrchestratorSetupRequest(BaseModel):
    projectID: str
    modelType: str
    system_prompt: Optional[str] = None
    dataset: Optional[str] = None
    custom_dataset: Optional[str] = None
    attackPack: Optional[str] = None
    attackSource: str
    parallelism: Optional[int] = None
    labels: Optional[List[str]] = None
    exclude: Optional[List[str]] = None
    include: Optional[List[str]] = None
    prompt_repeats: Optional[int] = None
    seed_prompts: Optional[str] = None
    perspective: Optional[str] = None
    tone: Optional[str] = None

    @model_validator(mode="after")  # type: ignore
    def check_parallelism(self):
        if self.parallelism and self.parallelism < 1:
            raise ValueError("Parallelism must be greater than 0.")

        return self


class OrchestratorSetupResponse(BaseModel):
    url: str
    group_id: str


class AttackModel(BaseModel):
    id: str
    submitted_at: str
    submitted_at_unix: float
    run_at: str
    run_at_unix: float
    state: Literal[0, 1, 2, -1]
    state_message: Literal["Completed", "Queued", "Running", "Failed"]
    runtime: float
    model: str
    dataset: str
    attack: str
    risk: int
    stacktrace: Optional[str]


class OrchestratorTestResponse(BaseModel):
    id: str
    mindgardModelName: str
    source: type_orchestrator_source
    createdAt: str
    attacks: List[AttackModel]
    isCompleted: bool
    hasFinished: bool
    risk: int
    test_url: str


class GetTestAttacksAttack(BaseModel):
    id: str
    attack_name: str
    status: Literal[-1, 0, 1, 2]
    total_events: int
    flagged_events: int


class GetTestAttacksTest(BaseModel):
    has_finished: bool
    id: str
    project_id: str
    flagged_events: int
    total_events: int


class GetTestAttacksItem(BaseModel):
    attack: GetTestAttacksAttack


class GetTestAttacksResponse(BaseModel):
    items: list[GetTestAttacksItem]
    test: GetTestAttacksTest
    raw: Optional[Dict[str, Any]]


class GetTestListTest(BaseModel):
    id: str
    project_id: str
    created_at: str
    flagged_events: int
    total_events: int


class GetTestListResponse(BaseModel):
    items: list[GetTestListTest]


def submit_sandbox_test(
    access_token: str,
    mindgard_model_name: str,
    post_request_function: type_post_request_function = api_post,
    get_request_function: type_get_request_function = api_get,
) -> GetTestAttacksResponse:
    url = f"{API_BASE}/assessments"
    post_body = {
        "mindgardModelName": mindgard_model_name
    }  # NOTE: this is the desired LLM model to be used in the sandbox test
    res = post_request_function(url, access_token, post_body)
    id = res.json().get("id", None)
    project_id = res.json().get("project_id", None)
    return get_test_by_id(
        test_id=id, project_id=project_id, access_token=access_token, request_function=get_request_function
    )


def get_tests(
    access_token: str, project_id: str, request_function: type_get_request_function = api_get
) -> GetTestListResponse:
    url = f"{API_BASE}/tests?project_id={project_id}"

    try:
        response = request_function(url, access_token)
        return GetTestListResponse(**response.json())
    except Exception as e:
        raise e


def get_test_by_id(
    project_id: str,
    test_id: str,
    access_token: str,
    request_function: type_get_request_function = api_get,
) -> GetTestAttacksResponse:
    try:
        response = request_function(f"{API_BASE}/tests/{test_id}/attacks?project_id={project_id}", access_token)
        data = response.json()
        return GetTestAttacksResponse(**response.json(), raw=data)

    except HTTPError as httpe:
        if httpe.response.status_code == 404:
            raise ValueError(f"Test with test_id {test_id=} not found!")
        else:
            raise httpe

    except Exception as e:
        raise e


def get_extra_config_from_env() -> Dict[str, Any]:
    config_str = os.getenv("MINDGARD_EXTRA_CONFIG")
    return cast(Dict[str, Any], json.loads(config_str)) if config_str else {}


def setup_orchestrator_webpubsub_request(
    request: OrchestratorSetupRequest,
    access_token: str,
    request_function: type_post_request_function = api_post,
) -> OrchestratorSetupResponse:
    url = f"{API_BASE}/tests/cli_init"

    try:
        payload = request.model_dump()
        extra_config = os.environ.get("MINDGARD_EXTRA_CONFIG", None)
        if extra_config is not None:
            payload["extraConfig"] = json.loads(extra_config)
        data = request_function(url, access_token, payload)
        url = data.json().get("url", None)
        groupId = data.json().get("groupId", None)
        if not url or not groupId:
            raise ValueError("Error: Invalid response, missing websocket credentials.")
        return OrchestratorSetupResponse(url=url, group_id=groupId)

    except MGHTTPError as e:
        raise ValueError(f"Error: {str(e.message)}")



from ..constants import ENABLE_DEV_FEATURES
from .types import _SubparserType


def add_create_command(subparsers: _SubparserType):
    create_parser = subparsers.add_parser("create", help="Create commands")
    create_subparsers = create_parser.add_subparsers(dest="create_command")
    add_create_dataset_command(create_subparsers)
    add_create_project_command(create_subparsers)
    return create_parser


def add_create_dataset_command(create_subparsers: _SubparserType):
    create_dataset_parser = create_subparsers.add_parser("dataset", help="Create a custom dataset for your test")
    create_dataset_parser.add_argument(
        "--seed-prompt",
        type=str,
        help='A seed prompt representing a policy, for which a dataset encouraging violations of the policy will be generated.\nFor example: "The model should never generate harmful, unethical, or illegal content."',
        required=True,
    )
    create_dataset_parser.add_argument(
        "--perspective",
        type=str,
        help="The perspective to use while generating the dataset. This skews the dataset generation towards asking the same question, but through a historical, cultural, etc lens that may subvert a target model.",
        choices=["nonspecific", "historical", "cultural", "scientific"],
        default="nonspecific",
        required=False,
    )
    create_dataset_parser.add_argument(
        "--tone",
        type=str,
        help="The tone to use for the questions in the dataset.",
        choices=["neutral", "forceful", "leading", "innocent", "corrigible", "indirect"],
        default="neutral",
        required=False,
    )
    create_dataset_parser.add_argument(
        "--output-filename",
        type=str,
        help="Name of the file the dataset will be stored in.",
        default="mindgard_custom_dataset.txt",
    )
    create_dataset_parser.add_argument(
        "--num-entries",
        type=int,
        help="Number of dataset entries to generate. Provided number is a goal, but the LLM may generate more or less than requested.",
        default=15,
    )
    return create_dataset_parser


def add_create_project_command(create_subparsers: _SubparserType):
    create_project_parser = create_subparsers.add_parser("project", help="Create a project for your test executions")
    create_project_parser.add_argument(
        "--name",
        type=str,
        help="Name of the project to create.",
        required=True,
    )



# Standard library imports
import logging
from typing import Callable, Dict, Literal, Optional, Type

# Third party imports
from requests import HTTPError, status_codes


class MGException(Exception):
    def __init__(self, message: str = "exception") -> None:
        super().__init__(self)
        self.message = message

    def __str__(self) -> str:
        return self.message


# Other Exceptions
class Uncontactable(MGException):
    pass


# HTTP-like Exceptions
class HTTPBaseError(MGException):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code: int = status_code
        self.status_message: str = "<unknown>"
        try:
            self.status_message = status_codes._codes[status_code][0]
        except:
            pass


class BadRequest(HTTPBaseError):
    pass


class Unauthorized(HTTPBaseError):
    pass


class Unauthenticated(HTTPBaseError):
    pass


class Forbidden(HTTPBaseError):
    pass


class NotFound(HTTPBaseError):
    pass


class Timeout(HTTPBaseError):
    pass


class UnprocessableEntity(HTTPBaseError):
    pass


class FailedDependency(HTTPBaseError):
    pass


class RateLimitOrInsufficientCredits(HTTPBaseError):
    pass


class InternalServerError(HTTPBaseError):
    pass


class ServiceUnavailable(HTTPBaseError):
    pass


class EmptyResponse(MGException):
    pass


class NotImplemented(MGException):
    pass


class GatewayTimeout(HTTPBaseError): ...


msg_bad_request = "LLM provider received message that couldn't be handled."
msg_unauthorized = "User is not authorized to access this LLM provider resource"
msg_forbidden = "User is known but does not have permission to access this resource"
msg_not_found = "This resource was not found."
msg_timeout = "Timed out while trying to access resource"
msg_unprocessable = "Entity sent to LLM could not be processed."
msg_failed_dependency = "Failed Dependency"
msg_rate_limit_or_insufficient_credits = "Rate Limit or Insufficient Credits"
msg_internal_server_error = "Internal Server Error"
msg_service_unavailable = "Service Unavailable"
msg_gateway_timeout = "The server, acting as a gateway or proxy, didn't receive a timely response from an upstream server it needed to access to complete the request"

_status_code_exception_map: Dict[int, HTTPBaseError] = {
    400: BadRequest(msg_bad_request, 400),
    401: Unauthorized(msg_unauthorized, 401),
    403: Forbidden(msg_forbidden, 403),
    404: NotFound(msg_not_found, 404),
    408: Timeout(msg_timeout, 408),
    422: UnprocessableEntity(msg_unprocessable, 422),
    424: FailedDependency(msg_failed_dependency, 424),
    429: RateLimitOrInsufficientCredits(msg_rate_limit_or_insufficient_credits, 429),
    500: InternalServerError(msg_internal_server_error, 500),
    503: ServiceUnavailable(msg_service_unavailable, 503),
    504: GatewayTimeout(msg_gateway_timeout, 504),
}


def status_code_to_exception(status_code: int, actual_error: Optional[HTTPError] = None) -> HTTPBaseError:
    return _status_code_exception_map.get(
        status_code, HTTPBaseError("An unexpected error occurred:" + str(actual_error.response), status_code)
    )


ErrorCode = Literal[
    "CouldNotContact",
    "ContentPolicy",
    "CLIError",
    "NotImplemented",
    "NoResponse",
    "RateLimited",
    "NetworkIssue",
    "MaxContextLength",
]


exceptions_to_cli_status_codes: Dict[Type[Exception], ErrorCode] = {
    Uncontactable: "CouldNotContact",
    BadRequest: "NoResponse",  # not sure about this, we don't handle 400 atm
    Unauthorized: "NoResponse",
    Forbidden: "NoResponse",
    NotFound: "NoResponse",
    Timeout: "NoResponse",
    UnprocessableEntity: "NoResponse",  # this is currently being handled as a rate limit issue for some reason
    FailedDependency: "NoResponse",
    RateLimitOrInsufficientCredits: "RateLimited",
    InternalServerError: "NoResponse",
    ServiceUnavailable: "NoResponse",
    NotImplemented: "NotImplemented",
    EmptyResponse: "NoResponse",
    GatewayTimeout: "NoResponse",
}


def handle_exception_callback(
    exception: Exception,
    handle_visual_exception_callback: Optional[Callable[[str], None]],
) -> ErrorCode:
    # TODO - come take a look at this
    error_code: ErrorCode = exceptions_to_cli_status_codes.get(type(exception), "CLIError")
    callback_text = str(exception)

    if handle_visual_exception_callback:
        handle_visual_exception_callback(callback_text)

    logging.debug(exception)
    return error_code
